
library(mvtnorm)
library(class)

# Normalize data (by subtracting Mean and Dividing by SD of the RefData
normalizeData = function(DataToNormalize, RefData){
  normalizedData = DataToNormalize
  for (n in 1:ncol(DataToNormalize)){
    normalizedData[,n] = (DataToNormalize[,n] - mean(RefData[,n])) / sd(RefData[,n]) 
  }
  return(normalizedData) 
}

# Calculate initial scaling factors R0s 
InitialRs = function (X, eta, S0s) { 
  K = length(S0s); R0s = rep(0, K)
  for (k in 1:K) {
    # force <12 to avoid s=exp(d) overflow
    R0s[k] = min((-log(S0s[k])) ^(1/eta) / max(IQR(X[ ,k]), 0.0000001), 12) }
  return (R0s)
}

Distance = function (Rs, x1, x2) {
  K = length(Rs)
  d = 0; for (k in 1:K) { d = d + (Rs[k]* (x2[k]-x1[k]))^2 }
  return (d^0.5)
}

# Calculate Similarity Score. 
SimilarityTrain = function(eta, Rs, Xtrain) {
  N1 = nrow(Xtrain); K = length(Rs); 
  S = matrix(0, nrow=N1, ncol=N1)
  for (i in 1:N1) { for (j in i:N1) {
    d2 = 0; for (k in 1:K) { d2 = d2 + (Rs[k]* (Xtrain[i,k]-Xtrain[j,k]))^2 }
    S[i,j] = exp(-d2^0.5) # avoid 0 for i=j
    S[j,i]=S[i,j] # Use symmetry to reduce 50% CPU time.
  } } # End j and i loops
  return (S)
} 

# Calculate Similarity Scores between training and test subjects. 
Similarity = function(eta, Rs, Xtrain, Xtest) {
  N1 = nrow(Xtrain); N2=nrow(Xtest); K = length(Rs)
  S = matrix(0, nrow=N2, ncol=N1)
  for (i in 1:N2) { for (j in 1:N1) {
    d2 = 0; for (k in 1:K) { d2 = d2 + (Rs[k]* (Xtest[i,k]-Xtrain[j,k]))^2 }
    S[i,j] = exp(-d2^0.5)
  } } # End j and i loops
  return (S)
} 

Weight = function(S) {
  N1= nrow(S); N2=ncol(S);
  W = matrix(0, nrow=N1, ncol=N2)
  for (i in 1:N1) {
    sum_S_row = sum(S[i, ])
    for (j in 1:N2) { W[i,j] = S[i,j] / max(sum_S_row,0.00000000001) } # avoid overflow
  } 
  return (W) 
} # End of Weight

PredictedY = function (W, Ytrain, Ytest) {
  # Calculate predicted outcome and Error
  OutObj = list()
  OutObj$pred_Y = W %*% Ytrain # For binary outcome, pred_y = prob of being 1.
  # Mean squared error for continuous outcome
  OutObj$MSE = mean((OutObj$pred_Y - Ytest)^2) 
  # Misclassification error, use the following linear instead:
  # OutObj$MSE = mean((round(OutObj$pred_Y) - Ytest)^2)   
  return (OutObj)
} # End of PredictionY

DerivativeE = function(eta, pred_Y, Rs, X, S, O) {
  # Derivatives if the loss function 
  N = nrow(X); K =length(Rs)
  der_S = matrix(0, nrow=K*N, ncol=N); der_W = matrix(0, nrow=K*N, ncol=N)
  dist = matrix(0, nrow=N, ncol=N); der_E = rep(0, K)
  for (i in 1:N) { for (j in i:N) {
    d2 = 0; for (k in 1:K) { d2 = d2 + (Rs[k]* (X[i,k]-X[j,k]))^2 }
    dist[i,j] =max(d2^0.5, 0.0000001) # avoid overflow der_d below
    dist[j,i]=dist[i,j]
  }} # End of i and j loops
  for (m in 1:K) { for (i in 1:N) { for (j in i:N) {
    der_d = (Rs[m]/dist[i,j]) * (X[i,m]-X[j,m]) ^2
    der_S[(m-1)*N+i, j] = -1 * S[i,j] * eta * (dist[i,j])^(eta-1) * der_d
    der_S[(m-1)*N+j, i] =der_S[(m-1)*N+i, j]
  } } } # End of j, i, and m loops
  # Weight Derivative
  for (m in 1:K) { for (i in 1:N) {
    sum_der_S = sum(der_S[(m-1)*N+i, ]); sumSi = sum(S[i, ])
    for (j in 1:N) { 
      der_W[(m-1)*N+i, j] = der_S[(m-1)*N+i, j] / sumSi - S[i,j]* sum_der_S /sumSi^2
    } } } # End of j, i, and m loops
  # Derivatives of E 
  for (m in 1:K) { for (i in 1:N) {
    err = 2/N *(pred_Y[i] - O[i])  # For mean squared error
    for (j in 1:N) { der_E[m] = der_E[m] +  err * O[j] * der_W[(m-1)*N+i, j] } 
  } } # End of I and m loops  
  return (der_E)
} # End of DerivativeE

Learning = function (LearningRate, Lamda, Rs, der_E) { 
  # Update scaling factors, Rs.
  K=length(Rs)
  der_lossFun = der_E+2*Lamda*Rs # derivatives of loss function
  Rs = Rs - LearningRate * der_lossFun
  # force Rs<25 to avoid S=exp(d) overflow
  for (m in 1:length(Rs)) { Rs[m] = min(max(0,Rs[m]),25) } 
  return (Rs)
} # End of Learning

# Training AI to obtain scaling factors, Rs ###
SBMLtrain = function(Epoch, S0, Lamda, LearningRate, eta, Xtrain, Ytrain) {
  TrainObj=list(); OutObj0= list(); OutObj= list()
  R0 = InitialRs(Xtrain, eta, S0); 
  S = SimilarityTrain(eta, R0, Xtrain) 
  OutObj0 = PredictedY (Weight(S), Ytrain, Ytrain)
  Rs=R0; OutObj =OutObj0 # in case Epoch =0 for no learning 
  TrainMSE0 = OutObj0$MSE
  preLoss = OutObj0$MSE+Lamda*sum(Rs^2)
  # Learning
  iter=0; 
  while (iter<Epoch) {
    preRs=Rs
    Rs = Learning(LearningRate, Lamda, Rs, DerivativeE(eta, OutObj0$pred_Y, Rs, Xtrain, S, Ytrain)) 
    OutObj = PredictedY (Weight(SimilarityTrain (eta, Rs, Xtrain)), Ytrain, Ytrain)
    iter=iter+1
    Loss = OutObj$MSE+Lamda*sum(Rs^2)
    if (Loss>preLoss) {Rs=preRs; iter=Epoch+1}
    preLoss=Loss
  }
  TrainObj$Rs = Rs; TrainObj$Y = OutObj$pred_Y; TrainObj$MSE = OutObj$MSE
  return( TrainObj)
}


### The following example with Continuous Outcome is based on just one training dataset and one testing dataset of size n=100 each. 
### In the actual evaluation of all ML methods in the paper, we used many training and testing datasets and compared the average MSE results.
### Therefore, the simulations/evaluations took much longer computing time.

# Example of Continuous Outcome
# set.seed(8372)
# covariance matrix for creating correlated X.
sigma = matrix(c(1, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0,
                 0.0, 1, 0.0, 0.0, 0.0, 0.0, 0.0,
                 0.6, 0.0, 1, 0.5, 0.0, 0.0, 0.0,
                 0.2, 0.0, 0.5, 1, 0.0, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 1, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 1, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1), ncol=7) 
mu=c(0,0,0,0,0,0,0) # 7 multivarite normal means
initS = rep(0.5, 7) # 7 intial similarities.
xTrain = rmvnorm(n=100, mean=mu, sigma=sigma) 
xTest = rmvnorm(n=100, mean=mu, sigma=sigma) 
# Construct example of outcome y that are not linearly related to x
Ytrain=xTrain[,1]+xTrain[,2]*xTrain[,3]+2*xTrain[,4]^2+xTrain[,5]
Ytest=xTest[,1]+xTest[,2]*xTest[,3]+xTest[,4]^2+xTest[,5]
# Data normalization
Xtrain = as.data.frame(normalizeData(xTrain, xTrain))
Xtest = as.data.frame(normalizeData(xTest, xTrain))
mySBMLtrain=SBMLtrain(Epoch=5, S0=initS, Lamda=-0.3, LearningRate=0.125, eta=1, Xtrain, Ytrain) 
myPred=PredictedY(Weight(Similarity(eta=1, mySBMLtrain$Rs, Xtrain, Xtest)),Ytrain, Ytest)
mse1=mean((Ytest-mean(Ytrain))^2) # mse of the data
mse2=myPred$MSE  # MSE using predicted Y.
mse1; mse2


### The following example with Binary Outcome is based on just one training dataset and one testing dataset of size n=100 each. 
### In the actual evaluation of all ML methods in the paper, we used many training and testing datasets and compared the average MSE results.
### Therefore, the simulations/evaluations took much longer computing time.

# Example of Binary Outcome
# set.seed(8373)
# covariance matrix for creating correlated X.
sigma = matrix(c(1, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0,
                 0.0, 1, 0.0, 0.0, 0.0, 0.0, 0.0,
                 0.6, 0.0, 1, 0.5, 0.0, 0.0, 0.0,
                 0.2, 0.0, 0.5, 1, 0.0, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 1, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 1, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1), ncol=7) 
mu=c(0,0,0,0,0,0,0) # 7 multivarite normal means
initS = rep(0.5, 7) # 7 intial similarities.
xTrain = rmvnorm(n=100, mean=mu, sigma=sigma) 
xTest = rmvnorm(n=100, mean=mu, sigma=sigma) 
# Construct y that are not linearly related to x, using sign function to make binary
Ytrain = (sign(xTrain[,1]+xTrain[,2]*xTrain[,3]+2*xTrain[,4]^2+xTrain[,5])+1)/2
Ytest = (sign(xTest[,1]+xTest[,2]*xTest[,3]+xTest[,4]^2+xTest[,5])+1)/2

# Data normalization
Xtrain = as.data.frame(normalizeData(xTrain, xTrain))
Xtest =  as.data.frame(normalizeData(xTest, xTrain))
initS = rep(0.5, 7)
mySBMLtrain=SBMLtrain(Epoch=15, S0=initS, Lamda=-0.5, LearningRate=0.125, eta=1, Xtrain, Ytrain) 
mySBMLpred=PredictedY(Weight(Similarity(eta=1, mySBMLtrain$Rs, Xtrain, Xtest)),Ytrain, Ytest)
mse3=mySBMLpred$MSE  # error in probability of classification
mse4=mean((Ytest-round(mySBMLpred$pred_Y))^2) # Misclassification Error with prob=0.5 as cutpoint
mse3; mse4


##### Preparation of Pima Indians Diabetes Dataset for Demo of SBML ##### 

library(mlbench)
data(PimaIndiansDiabetes)
ntrain = 5; ntest = 3; lamdaVal = 1; EpochN = 0;  
myXvars = c("pressure", "age")  
myYvars = c("glucose")  
myvars = c(myXvars, myYvars)

### Trainset - Pima Indians Diabetes (PID)
myTrainPID = PimaIndiansDiabetes[1:ntrain, myvars]
xTrain = myTrainPID[, myXvars]
yTrain = myTrainPID[, myYvars]

### Testset - Pima Indians Diabetes (PID)
myTestPID = PimaIndiansDiabetes[(ntrain+1):(ntrain+ntest), myvars]
xTest = myTestPID[, myXvars]
yTest = myTestPID[, myYvars]

head(myTrainPID)
head(myTestPID)
xTrain
initS
eta 
R0s

PredictedY(W, yTrain, yTrain) 
mySBMLtrain = SBMLtrain(Epoch=EpochN, S0s=initS, Lamda=lamdaVal, LearningRate=LRval, eta=1, xTrain, yTrain)
mySBMLtrain ## <- normalized Xs, raw Ys
mySBMLtrain ## <- Raw Xs and Ys (i.e., NOT normalized data)
cbind(PredictedY(W, Ytrain, Ytrain)$pred_Y, yTrain)
sum((PredictedY(W, Ytrain, Ytrain)$pred_Y - yTrain)^2) / 5

cbind(PredictedY(W, Ytrain, Ytrain)$pred_Y, yTrain) #<-normalized Xs, raw Ys
yTrain

sum((PredictedY(W, Ytrain, Ytrain)$pred_Y - yTrain)^2) / 5
predY
predY = PredictedY(W=Weight(Similarity(eta=eta, mySBMLtrain$Rs, xTrain, xTest)), Ytrain=yTrain, Ytest=yTest)
predY

cbind(PredictedY(W, yTrain, yTest)$pred_Y, yTest)
cbind(PredictedY(W, yTrain, yTest)$pred_Y, yTest) #<-normalized Xs, raw Ys

cbind(PredictedY(W, yTrain, yTest)$pred_Y, yTest)

#####
